"""
Encapsulate the period of a feature, defined by a unit of time
measure, a start and an end
"""

from datetime import date

from jstark.period_unit_of_measure import PeriodUnitOfMeasure
from .exceptions import FeaturePeriodEndGreaterThanStartError


class FeaturePeriod:
    """
    Encapsulate the period of a feature, defined by a unit of time
    measure, a start and an end
    """

    def __init__(
        self, period_unit_of_measure: PeriodUnitOfMeasure, start: int, end: int
    ) -> None:
        if not isinstance(period_unit_of_measure, PeriodUnitOfMeasure):
            raise TypeError(
                (
                    "period_unit_of_measure needs to be of type "
                    + f"PeriodUnitOfMeasure, not {type(period_unit_of_measure)}"
                )
            )
        if end > start:
            raise FeaturePeriodEndGreaterThanStartError(start=start, end=end)
        self.__period_unit_of_measure = period_unit_of_measure
        self.__start = start
        self.__end = end

    @property
    def start(self) -> int:
        "Number of periods ago that the FeaturePeriod begins at"
        return self.__start

    @property
    def end(self) -> int:
        "Number of periods ago that the FeaturePeriod ends at"
        return self.__end

    @property
    def period_unit_of_measure(self) -> PeriodUnitOfMeasure:
        "Period unit of measure"
        return self.__period_unit_of_measure

    @property
    def mnemonic(self) -> str:
        "Mnemonic for the feature period"
        return f"{self.start}{self.period_unit_of_measure.value}{self.end}"

    @property
    def description(self) -> str:
        """Description of the feature period

        Pretty sure this will change in time, but this initial implementation
        will do for now

        Returns:
            str: description
        """
        return (
            f"Between {self.start} and {self.end} "
            + f"{self.period_unit_of_measure.name.lower()}s ago"
        )

    @property
    def number_of_periods(self) -> int:
        "Number of periods between start and end (inclusive)"
        return self.start - self.end + 1

    def __str__(self) -> str:
        return self.description

    def __repr__(self) -> str:
        return (
            f"FeaturePeriod("
            f"period_unit_of_measure={self.period_unit_of_measure}, "
            f"start={self.start}, end={self.end})"
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FeaturePeriod):
            return False
        return (
            self.period_unit_of_measure == other.period_unit_of_measure
            and self.start == other.start
            and self.end == other.end
        )

    def __hash__(self) -> int:
        return hash((self.period_unit_of_measure, self.start, self.end))


TODAY = {
    "as_at": date.today(),
    "feature_periods": [FeaturePeriod(PeriodUnitOfMeasure.DAY, 0, 0)],
}
YESTERDAY = {
    "as_at": date.today(),
    "feature_periods": [FeaturePeriod(PeriodUnitOfMeasure.DAY, 1, 1)],
}
THIS_WEEK = {
    "as_at": date.today(),
    "feature_periods": [FeaturePeriod(PeriodUnitOfMeasure.WEEK, 0, 0)],
}
LAST_WEEK = {
    "as_at": date.today(),
    "feature_periods": [FeaturePeriod(PeriodUnitOfMeasure.WEEK, 1, 1)],
}
THIS_MONTH = {
    "as_at": date.today(),
    "feature_periods": [FeaturePeriod(PeriodUnitOfMeasure.MONTH, 0, 0)],
}
LAST_MONTH = {
    "as_at": date.today(),
    "feature_periods": [FeaturePeriod(PeriodUnitOfMeasure.MONTH, 1, 1)],
}
THIS_QUARTER = {
    "as_at": date.today(),
    "feature_periods": [FeaturePeriod(PeriodUnitOfMeasure.QUARTER, 0, 0)],
}
LAST_QUARTER = {
    "as_at": date.today(),
    "feature_periods": [FeaturePeriod(PeriodUnitOfMeasure.QUARTER, 1, 1)],
}
THIS_YEAR = {
    "as_at": date.today(),
    "feature_periods": [FeaturePeriod(PeriodUnitOfMeasure.YEAR, 0, 0)],
}
LAST_YEAR = {
    "as_at": date.today(),
    "feature_periods": [FeaturePeriod(PeriodUnitOfMeasure.YEAR, 1, 1)],
}
current_year = date.today().year
current_month = date.today().month
match current_month:
    case 1 | 2 | 3:
        quarters_this_year = [1]
    case 4 | 5 | 6:
        quarters_this_year = [1, 2]
    case 7 | 8 | 9:
        quarters_this_year = [1, 2, 3]
    case _:  # all other months:
        quarters_this_year = [1, 2, 3, 4]
ALL_MONTHS_LAST_YEAR = {
    "as_at": date(current_year, 1, 1),
    "feature_periods": [
        FeaturePeriod(PeriodUnitOfMeasure.MONTH, i, i) for i in range(1, 13)
    ],
}
ALL_MONTHS_THIS_YEAR = {
    "as_at": date.today(),
    "feature_periods": [
        FeaturePeriod(PeriodUnitOfMeasure.MONTH, i, i) for i in range(current_month)
    ],
}
ALL_QUARTERS_LAST_YEAR = {
    "as_at": date(current_year, 1, 1),
    "feature_periods": [
        FeaturePeriod(PeriodUnitOfMeasure.QUARTER, i, i) for i in range(1, 5)
    ],
}
ALL_QUARTERS_THIS_YEAR = {
    "as_at": date.today(),
    "feature_periods": [
        FeaturePeriod(PeriodUnitOfMeasure.QUARTER, i - 1, i - 1)
        for i in quarters_this_year
    ],
}
THIS_MONTH_VS_ONE_YEAR_PRIOR = {
    "as_at": date.today(),
    "feature_periods": [
        FeaturePeriod(PeriodUnitOfMeasure.MONTH, 0, 0),
        FeaturePeriod(PeriodUnitOfMeasure.MONTH, 12, 12),
    ],
}
LAST_MONTH_VS_ONE_YEAR_PRIOR = {
    "as_at": date.today(),
    "feature_periods": [
        FeaturePeriod(PeriodUnitOfMeasure.MONTH, 1, 1),
        FeaturePeriod(PeriodUnitOfMeasure.MONTH, 13, 13),
    ],
}
THIS_QUARTER_VS_ONE_YEAR_PRIOR = {
    "as_at": date.today(),
    "feature_periods": [
        FeaturePeriod(PeriodUnitOfMeasure.QUARTER, 0, 0),
        FeaturePeriod(PeriodUnitOfMeasure.QUARTER, 4, 4),
    ],
}

LAST_QUARTER_VS_ONE_YEAR_PRIOR = {
    "as_at": date.today(),
    "feature_periods": [
        FeaturePeriod(PeriodUnitOfMeasure.QUARTER, 1, 1),
        FeaturePeriod(PeriodUnitOfMeasure.QUARTER, 5, 5),
    ],
}

LAST_FIVE_YEARS = {
    "as_at": date.today(),
    "feature_periods": [
        FeaturePeriod(PeriodUnitOfMeasure.YEAR, i, i) for i in range(1, 6)
    ],
}



"""AvgQuantityPerBasket feature"""

import pyspark.sql.functions as f
from pyspark.sql import Column

from jstark.features.feature import DerivedFeature
from jstark.features.quantity import Quantity
from .order_count import OrderCount


class AvgQuantityPerOrder(DerivedFeature):
    def column_expression(self) -> Column:
        return f.try_divide(
            Quantity(
                self.as_at,
                self.feature_period,
                first_day_of_week=self._first_day_of_week,
            ).column,
            OrderCount(
                self.as_at,
                self.feature_period,
                first_day_of_week=self._first_day_of_week,
            ).column,
        )

    @property
    def description_subject(self) -> str:
        return "Average Quantity per Order"

    @property
    def commentary(self) -> str:
        return (
            "Total Quantity divided by the number of orders. "
            + "Very useful to know how large, on average, each order is."
        )

    def default_value(self) -> Column:
        return f.lit(None)



"""ProductCount feature"""

import pyspark.sql.functions as f
from pyspark.sql import Column

from jstark.features.distinctcount_feature import DistinctCount


class ProductCount(DistinctCount):
    def column_expression(self) -> Column:
        return f.col("Product")

    @property
    def description_subject(self) -> str:
        return "Distinct count of Products"

    @property
    def commentary(self) -> str:
        return (
            "The number of products. Typically the dataframe supplied "
            + "to this feature will have many transactions for many baskets with "
            + "many products bought, this feature allows you to determine how many "
            + "disinct products were bought in those baskets."
        )



"""MinGrossSpen feature"""

import pyspark.sql.functions as f
from pyspark.sql import Column

from jstark.features.min_feature import Min


class MinGrossSpend(Min):
    def column_expression(self) -> Column:
        return f.col("GrossSpend")

    @property
    def description_subject(self) -> str:
        return "Minimum GrossSpend value"



"""MinNetSpend feature"""

import pyspark.sql.functions as f
from pyspark.sql import Column

from jstark.features.min_feature import Min


class MinNetSpend(Min):
    def column_expression(self) -> Column:
        return f.col("NetSpend")

    @property
    def description_subject(self) -> str:
        return "Minimum of NetSpend value"



import random
from functools import cached_property
import uuid
from datetime import date
from typing import Any, Iterable
from decimal import Decimal

from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.types import (
    DecimalType,
    IntegerType,
    StringType,
    StructField,
    StructType,
    TimestampType,
)
from faker import Faker
from faker.providers import DynamicProvider


class FakeGroceryTransactions:
    def __init__(self, seed: int | None = None, number_of_baskets: int = 1000):
        self.seed = seed
        self.number_of_baskets = number_of_baskets

    @property
    def transactions_schema(self) -> StructType:
        return StructType(
            [
                StructField("Timestamp", TimestampType(), True),
                StructField("Customer", StringType(), True),
                StructField("Store", StringType(), True),
                StructField("Channel", StringType(), True),
                StructField("Product", StringType(), True),
                StructField("Quantity", IntegerType(), True),
                StructField("Basket", StringType(), True),
                StructField("GrossSpend", DecimalType(10, 2), True),
                StructField("NetSpend", DecimalType(10, 2), True),
                StructField("Discount", DecimalType(10, 2), True),
            ]
        )

    @staticmethod
    def flatten_transactions(transactions: list[Any]) -> Iterable[dict[str, Any]]:
        return [
            {
                "Customer": d["Customer"],
                "Store": d["Store"],
                "Basket": d["Basket"],
                "Channel": d["Channel"],
                "Timestamp": d["Timestamp"],
                **d2,
            }
            for d in transactions
            for d2 in d["items"]
        ]

    @cached_property
    def df(self) -> DataFrame:
        stores_provider = DynamicProvider(
            provider_name="store",
            elements=["Hammersmith", "Ealing", "Richmond", "Twickenham", "Staines"],
        )
        products_provider = DynamicProvider(
            provider_name="product",
            elements=[
                ("Custard Creams", "Ambient", 1.00),
                ("Carrots", "Fresh", 0.69),
                ("Cheddar", "Dairy", 3.43),
                ("Ice Cream", "Frozen", 5.32),
                ("Milk 1l", "Dairy", 1.70),
                ("Apples", "Fresh", 1.50),
                ("Beer", "BWS", 8.50),
                ("Wine", "BWS", 7.50),
                ("Yoghurt", "Dairy", 0.99),
                ("Bananas", "Fresh", 0.79),
                ("Nappies", "Baby", 15.00),
                ("Baby formula", "Baby", 5.99),
                ("Whisky", "BWS", 23.00),
            ],
        )
        channels_provider = DynamicProvider(
            provider_name="channel",
            elements=["Instore", "Online", "Click and Collect"],
        )

        fake = Faker()
        if self.seed:
            Faker.seed(self.seed)
        fake.add_provider(stores_provider)
        fake.add_provider(channels_provider)

        products_fake = Faker()
        products_fake.add_provider(products_provider)

        transactions = []

        possible_quantities = [1, 2, 3, 4, 5]
        if self.seed:
            random.seed(self.seed)
        quantities = random.choices(
            possible_quantities,
            weights=[100, 40, 20, 10, 8],
            k=self.number_of_baskets * len(products_provider.elements),
        )
        for basket in range(self.number_of_baskets):
            items = []
            if self.seed:
                random.seed(self.seed)
            for item in range(random.randint(1, len(products_provider.elements))):
                p = products_fake.unique.product()
                quantity = quantities[(basket * len(possible_quantities)) + item]
                gross_spend: float = round(p[2] * quantity, 2)
                net_spend = round((gross_spend * random.random()), 2)
                discount = round((gross_spend - net_spend) * random.random(), 2)
                items.append(
                    {
                        "Product": p[0],
                        "ProductCategory": p[1],
                        "Quantity": quantity,
                        "GrossSpend": Decimal(gross_spend),
                        "NetSpend": Decimal(net_spend),
                        "Discount": Decimal(discount),
                    }
                )
            transactions.append(
                {
                    "Customer": fake.name(),
                    "Homecity": fake.city(),
                    "Store": fake.store(),
                    "Timestamp": fake.date_time_between(
                        start_date=date(2021, 1, 1), end_date=date(2021, 12, 31)
                    ),
                    "Basket": str(uuid.uuid4()),
                    "Channel": fake.channel(),
                    "items": items,
                }
            )
            products_fake.unique.clear()

        flattened_transactions = self.flatten_transactions(transactions)
        spark = SparkSession.builder.getOrCreate()
        return spark.createDataFrame(
            flattened_transactions,
            schema=self.transactions_schema,  # type: ignore
        )



from datetime import date, timedelta
import pytest


from jstark.exceptions import AsAtIsNotADate
from jstark.features.gross_spend import GrossSpend
from jstark.feature_period import FeaturePeriod, PeriodUnitOfMeasure


def test_errors_if_as_at_is_not_a_date():
    with pytest.raises(AsAtIsNotADate) as exc_info:
        GrossSpend(
            as_at="2000-01-01",  # type: ignore
            feature_period=FeaturePeriod(PeriodUnitOfMeasure.DAY, 2, 1),
        )
    assert str(exc_info.value) == "as_at value must be of type date"


def test_as_at_is_immutable():
    """
    We don't want as_at to be changed. This is a choice we have made
    so let's assert that behaviour.
    """
    as_at = date(2022, 11, 30)
    _0d0 = GrossSpend(as_at, FeaturePeriod(PeriodUnitOfMeasure.DAY, 0, 0))
    assert _0d0.as_at == as_at
    with pytest.raises(AttributeError) as exc_info:
        _0d0.as_at = date(2022, 11, 29)
    assert (
        str(exc_info.value) == "property 'as_at' of 'GrossSpend' object has no setter"
    )


def test_start_date_days():
    as_at = date(2022, 11, 30)
    _0d0 = GrossSpend(as_at, FeaturePeriod(PeriodUnitOfMeasure.DAY, 0, 0))
    assert _0d0.start_date == as_at
    assert _0d0.end_date == as_at
    _2d0 = GrossSpend(as_at, FeaturePeriod(PeriodUnitOfMeasure.DAY, 2, 0))
    assert _2d0.start_date == date(2022, 11, 28)
    assert _2d0.end_date == as_at


def test_start_and_end_date_when_period_unit_of_measure_is_weeks():
    first_day_of_week = date(2022, 11, 28)  # this is a Monday
    for i in range(7):
        as_at = first_day_of_week + timedelta(days=i)
        # all days in the same week should return the same results
        _0w0 = GrossSpend(as_at, FeaturePeriod(PeriodUnitOfMeasure.WEEK, 0, 0))
        _1w0 = GrossSpend(as_at, FeaturePeriod(PeriodUnitOfMeasure.WEEK, 1, 0))
        _1w1 = GrossSpend(as_at, FeaturePeriod(PeriodUnitOfMeasure.WEEK, 1, 1))
        _2w0 = GrossSpend(as_at, FeaturePeriod(PeriodUnitOfMeasure.WEEK, 2, 0))
        _2w1 = GrossSpend(as_at, FeaturePeriod(PeriodUnitOfMeasure.WEEK, 2, 1))
        assert _0w0.start_date == date(2022, 11, 28)
        assert _0w0.end_date == as_at
        assert _1w0.start_date == date(2022, 11, 21)
        assert _1w0.end_date == as_at
        assert _1w1.start_date == date(2022, 11, 21)
        assert _1w1.end_date == date(2022, 11, 27)
        assert _2w0.start_date == date(2022, 11, 14)
        assert _2w0.end_date == as_at
        assert _2w1.start_date == date(2022, 11, 14)
        assert _2w1.end_date == date(2022, 11, 27)


def test_start_and_end_date_when_period_unit_of_measure_is_months():
    first_day_of_month = date(2022, 11, 1)
    for i in range(30):
        as_at = first_day_of_month + timedelta(days=i)
        _0m0 = GrossSpend(as_at, FeaturePeriod(PeriodUnitOfMeasure.MONTH, 0, 0))
        _1m0 = GrossSpend(as_at, FeaturePeriod(PeriodUnitOfMeasure.MONTH, 1, 0))
        _1m1 = GrossSpend(as_at, FeaturePeriod(PeriodUnitOfMeasure.MONTH, 1, 1))
        _12m0 = GrossSpend(as_at, FeaturePeriod(PeriodUnitOfMeasure.MONTH, 12, 0))
        _12m12 = GrossSpend(as_at, FeaturePeriod(PeriodUnitOfMeasure.MONTH, 12, 12))
        assert _0m0.start_date == first_day_of_month
        assert _0m0.end_date == as_at
        assert _1m0.start_date == date(2022, 10, 1)
        assert _0m0.end_date == as_at
        assert _1m1.start_date == date(2022, 10, 1)
        assert _1m1.end_date == date(2022, 10, 31)
        assert _12m0.start_date == date(2021, 11, 1)
        assert _12m0.end_date == as_at
        assert _12m12.start_date == date(2021, 11, 1)
        assert _12m12.end_date == date(2021, 11, 30)


def test_start_and_end_date_when_period_unit_of_measure_is_quarter():
    first_day_of_quarter = date(2022, 10, 1)
    for i in range(92):
        as_at = first_day_of_quarter + timedelta(days=i)
        _0q0 = GrossSpend(as_at, FeaturePeriod(PeriodUnitOfMeasure.QUARTER, 0, 0))
        _1q1 = GrossSpend(as_at, FeaturePeriod(PeriodUnitOfMeasure.QUARTER, 1, 1))
        _2q2 = GrossSpend(as_at, FeaturePeriod(PeriodUnitOfMeasure.QUARTER, 2, 2))
        _3q3 = GrossSpend(as_at, FeaturePeriod(PeriodUnitOfMeasure.QUARTER, 3, 3))
        _4q4 = GrossSpend(as_at, FeaturePeriod(PeriodUnitOfMeasure.QUARTER, 4, 4))
        _4q0 = GrossSpend(as_at, FeaturePeriod(PeriodUnitOfMeasure.QUARTER, 4, 0))
        assert _0q0.start_date == date(2022, 10, 1)
        assert _0q0.end_date == as_at
        assert _1q1.start_date == date(2022, 7, 1)
        assert _1q1.end_date == date(2022, 9, 30)
        assert _2q2.start_date == date(2022, 4, 1)
        assert _2q2.end_date == date(2022, 6, 30)
        assert _3q3.start_date == date(2022, 1, 1)
        assert _3q3.end_date == date(2022, 3, 31)
        assert _4q4.start_date == date(2021, 10, 1)
        assert _4q4.end_date == date(2021, 12, 31)
        assert _4q0.start_date == date(2021, 10, 1)
        assert _4q0.end_date == as_at


def test_start_and_end_date_when_period_unit_of_measure_is_year():
    first_day_of_year = date(2022, 1, 1)
    for i in range(365):
        as_at = first_day_of_year + timedelta(days=i)
        _0y0 = GrossSpend(as_at, FeaturePeriod(PeriodUnitOfMeasure.YEAR, 0, 0))
        _1y1 = GrossSpend(as_at, FeaturePeriod(PeriodUnitOfMeasure.YEAR, 1, 1))
        assert _0y0.start_date == date(2022, 1, 1)
        assert _0y0.end_date == as_at
        assert _1y1.start_date == date(2021, 1, 1)
        assert _1y1.end_date == date(2021, 12, 31)


# def test_metadata():
#     f = GrossSpend(as_at, FeaturePeriod(PeriodUnitOfMeasure.YEAR, 0, 0))
#     f.



from datetime import date
import pytest
from jstark.features.first_and_last_date_of_period import FirstAndLastDateOfPeriod


def test_invalid_first_day_of_week():
    assert pytest.raises(
        ValueError, FirstAndLastDateOfPeriod, date(2022, 12, 8), first_day_of_week="Sat"
    )


def test_first_date_in_week_for_default_first_day_of_week():
    assert FirstAndLastDateOfPeriod(date(2022, 12, 8)).first_date_in_week == date(
        2022, 12, 5
    )


def test_last_date_in_week_for_default_first_day_of_week():
    assert FirstAndLastDateOfPeriod(date(2022, 12, 8)).last_date_in_week == date(
        2022, 12, 11
    )


def test_first_date_in_week_with_saturday_as_first_day_of_week():
    assert FirstAndLastDateOfPeriod(
        date(2022, 12, 8), first_day_of_week="Saturday"
    ).first_date_in_week == date(2022, 12, 3)


def test_last_date_in_week_with_saturday_as_first_day_of_week():
    last_date_in_week = FirstAndLastDateOfPeriod(
        date(2022, 12, 8), first_day_of_week="Saturday"
    ).last_date_in_week
    assert last_date_in_week == date(2022, 12, 9)


def test_first_date_in_week_with_sunday_as_first_day_of_week():
    assert FirstAndLastDateOfPeriod(
        date(2022, 12, 8), first_day_of_week="Sunday"
    ).first_date_in_week == date(2022, 12, 4)


def test_first_date_in_week_with_monday_as_first_day_of_week():
    assert FirstAndLastDateOfPeriod(
        date(2022, 12, 8), first_day_of_week="Monday"
    ).first_date_in_week == date(2022, 12, 5)


def test_first_date_in_week_with_tuesday_as_first_day_of_week():
    assert FirstAndLastDateOfPeriod(
        date(2022, 12, 8), first_day_of_week="Tuesday"
    ).first_date_in_week == date(2022, 12, 6)


def test_first_date_in_week_with_wednesday_as_first_day_of_week():
    assert FirstAndLastDateOfPeriod(
        date(2022, 12, 8), first_day_of_week="Wednesday"
    ).first_date_in_week == date(2022, 12, 7)


def test_first_date_in_week_with_thursday_as_first_day_of_week():
    assert FirstAndLastDateOfPeriod(
        date(2022, 12, 8), first_day_of_week="Thursday"
    ).first_date_in_week == date(2022, 12, 8)


def test_first_date_in_week_with_friday_as_first_day_of_week():
    assert FirstAndLastDateOfPeriod(
        date(2022, 12, 8), first_day_of_week="Friday"
    ).first_date_in_week == date(2022, 12, 2)


def test_first_date_in_month():
    assert FirstAndLastDateOfPeriod(date(2022, 12, 8)).first_date_in_month == date(
        2022, 12, 1
    )


def test_last_date_in_month():
    assert FirstAndLastDateOfPeriod(date(2022, 12, 8)).last_date_in_month == date(
        2022, 12, 31
    )


def test_first_date_in_quarter():
    assert FirstAndLastDateOfPeriod(date(2022, 12, 8)).first_date_in_quarter == date(
        2022, 10, 1
    )


def test_last_date_in_quarter():
    assert FirstAndLastDateOfPeriod(date(2022, 12, 8)).last_date_in_quarter == date(
        2022, 12, 31
    )


def test_first_date_in_year():
    assert FirstAndLastDateOfPeriod(date(2022, 12, 8)).first_date_in_year == date(
        2022, 1, 1
    )


def test_last_date_in_year():
    assert FirstAndLastDateOfPeriod(date(2022, 12, 8)).last_date_in_year == date(
        2022, 12, 31
    )
